"""Entry point.

"""
from os.path import join, expanduser

from cml.usecases.query import (
    PreprocessUsecase,
    KnowledgeSearchUsecase,
    CreateConstructorUsecase,
    FeatureSelectionUsecase,
    ReconstructionUsecase
)
from cml.ports.source_adapters import PandasAdapter
from cml.ports.ml_adapter import (
    ConstructionClusteringMLModel,
    FilterMethod,
    EmbeddedMethod
)
from cml.shared.settings import Settings
from cml.shared.settings import specific_settings_factory, read_settings
from cml.shared.request import (
    PreprocessRequest,
    KnowledgeSearchRequest,
    CreateConstructorRequest,
    FeatureSelectionRequest,
    ReconstructionRequest
)
from cml.ports.ml_adapter import (
    KernelDensityEstimator,
    find_relative_extrema,
    Autoencoder
)


__all__ = (
    "load_settings",
    "get_settings",
    "get_data_source",
    "construction",
    "reconstruction",
    "reconstruction",
    "search_knowledge",
    "visualizer",
    "save_knowledge",
    "load_knowledge",
    "feature_selection"
)


def default_path(func):
    def wrapper(path: str = None):
        if not path:
            path = join(expanduser("~"), ".cml", "settings.ini")

        try:
            func(path)
        except FileNotFoundError as e:
            # TODO (dmt):Provide proper exception handling!
            pass
    return wrapper


@default_path
def load_settings(path: str):
    read_settings(path)


def get_settings():
    return Settings


def get_data_source(path: str = None):
    general_settings = specific_settings_factory("general")
    preprocessing_settings = specific_settings_factory("preprocessing")
    block_processing_settings = specific_settings_factory("block_processing")
    if path:
        general_settings.input_file = path
    source_adapter = PandasAdapter.read_csv_data(general_settings.input_file)
    density_estimator = KernelDensityEstimator()
    relative_extrema = find_relative_extrema

    preprocessing_req = PreprocessRequest(source_adapter,
                                          preprocessing_settings,
                                          block_processing_settings,
                                          density_estimator,
                                          relative_extrema)
    preprocessing_usecase = PreprocessUsecase()
    return preprocessing_usecase.execute(preprocessing_req)


def construction(construct_type, *args, **kwargs):
    construction_settings = specific_settings_factory("construction")
    construction_settings.construct_type = construct_type

    if construct_type == "conceptual":
        ml_models = [ConstructionClusteringMLModel(raw_model)
                     for raw_model in args]

    elif construct_type == "procedural":
        ml_models = [Autoencoder]

    else:
        raise Exception("Operation not supported!")

    create_constructor_req = CreateConstructorRequest(construction_settings,
                                                      ml_models)
    create_construction_usecase = CreateConstructorUsecase()
    return create_construction_usecase.execute(create_constructor_req)


def feature_selection(filter_ml_model, embedded_ml_model):
    filter_method = FilterMethod(filter_ml_model)
    embedded_method = EmbeddedMethod(embedded_ml_model)
    settings = specific_settings_factory("feature_selection")
    feature_selection_req = FeatureSelectionRequest(filter_method,
                                                    embedded_method,
                                                    settings)
    feature_selection_usecase = FeatureSelectionUsecase()
    return feature_selection_usecase.execute(feature_selection_req)


def reconstruction(*args, **kwargs):
    settings = specific_settings_factory("reconstruction")
    reconstruction_req = ReconstructionRequest(settings)
    reconstruction_usecase = ReconstructionUsecase()
    return reconstruction_usecase.execute(reconstruction_req)


def search_knowledge(constructor, feature_selector, reconstructor, data_source,
                     stdout=False):
    deconstruction_settings = specific_settings_factory("deconstruction")
    knowledge_search_req = KnowledgeSearchRequest(constructor,
                                                  feature_selector,
                                                  reconstructor,
                                                  data_source,
                                                  deconstruction_settings,
                                                  stdout)
    knowledge_search_usecase = KnowledgeSearchUsecase()
    return knowledge_search_usecase.execute(knowledge_search_req)


def visualizer():
    pass


def save_knowledge():
    pass


def load_knowledge():
    pass
